/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_ASSETS_MANAGER_HPP
#define FGGL_ASSETS_MANAGER_HPP

#include <string_view>
#include <map>
#include <functional>
#include <memory>

#include "fggl/debug/logging.hpp"
#include "fggl/assets/types.hpp"
#include "fggl/util/safety.hpp"

namespace fggl::assets {

	struct AssetBox {
		virtual ~AssetBox() = default;
		virtual void release() = 0;
	};

	template<typename T>
	struct AssetBoxT : public AssetBox {
		T* asset = nullptr;

		inline void release() override {
			asset = nullptr;
		}
	};

	class AssetManager {
		public:
			constexpr const static modules::ModuleService service = modules::make_service("fggl::assets::Manager");

			AssetManager() = default;
			virtual ~AssetManager() = default;

			// no move, no copy.
			AssetManager(const AssetManager &) = delete;
			AssetManager &operator=(const AssetManager &) = delete;
			AssetManager(AssetManager &&) = delete;
			AssetManager &operator=(AssetManager &&) = delete;

			template<typename T>
			T* get(const AssetGUID &guid) const {
				try {
					const auto &assetRecord = m_registry.at(guid);
					std::shared_ptr<AssetBoxT<T>> casted = std::dynamic_pointer_cast<AssetBoxT<T>>(assetRecord);
					if ( casted == nullptr ) {
						debug::error("Asset type requested did not match loaded asset type!");
						return nullptr;
					}
					return casted->asset;
				} catch (std::out_of_range& e) {
					return nullptr;
				}
			}

			template<typename T>
			T* set(const AssetGUID &guid, T* assetRef) {
				auto ptr = std::make_shared<AssetBoxT<T>>();
				ptr->asset = assetRef;

				m_registry[guid] = ptr;
				return (*ptr).asset;
			}

			inline void require(const AssetGUID &guid) {
				//m_registry.at(guid).refCount++;
			}

			inline bool has(const AssetGUID &guid) {
				return m_registry.contains(guid);
			}

			inline void release(const AssetGUID &guid) {
				//m_registry.at(guid).refCount--;
			}

		private:
			std::map<AssetGUID, std::shared_ptr<AssetBox>> m_registry;
	};

} // namespace fggl::assets

#endif //FGGL_ASSETS_MANAGER_HPP
