/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 22/10/22.
//

#ifndef FGGL_GFX_OGL4_MESHES_HPP
#define FGGL_GFX_OGL4_MESHES_HPP

#include "fggl/mesh/components.hpp"
#include "fggl/entity/entity.hpp"

#include "fggl/gfx/ogl/types.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

namespace fggl::gfx::ogl4 {

	struct DrawType {
		ogl::Primitive mode;
		std::size_t restartIndex = ogl::NO_RESTART_IDX;
	};

	struct MeshData {
		std::shared_ptr<ogl::VertexArray> vao;
		std::shared_ptr<ogl::ElementBuffer> elements;
		std::shared_ptr<ogl::ArrayBuffer> vertexData;
		std::size_t elementCount;
		DrawType drawInfo;

		void draw() const;
	};

	struct StaticMesh {
		constexpr static auto name = "StaticMultiMesh";

		std::shared_ptr<ogl::Shader> pipeline;
		MeshData mesh;

		inline void draw() const {
			mesh.draw();
		}
	};

	struct StaticMultiMesh {
		constexpr static auto name = "StaticMultiMesh";

		std::shared_ptr<ogl::Shader> pipeline;
		std::vector<MeshData> meshes;

		void draw() const;
	};

	void setup_material(const std::shared_ptr<ogl::Shader>& shader, const PhongMaterial* material);
	void setup_lighting(const std::shared_ptr<ogl::Shader>& shader, const math::mat4& viewMatrix, const math::Transform& camTransform, const math::Transform& transform, math::vec3 lightPos);

	template<typename T>
	void forward_pass(const entity::EntityID& camera, const fggl::entity::EntityManager& world) {

		// enable required OpenGL state
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		// enable depth testing
		glEnable(GL_DEPTH_TEST);

		// set-up camera matrices
		const auto &camTransform = world.get<fggl::math::Transform>(camera);
		const auto &camComp = world.get<fggl::gfx::Camera>(camera);

		const math::mat4 projectionMatrix = camComp.perspective();
		const math::mat4 viewMatrix = glm::lookAt(camTransform.origin(), camComp.target, camTransform.up());

		// TODO lighting needs to not be this...
		math::vec3 lightPos{0.0F, 10.0F, 0.0F};

		std::shared_ptr<ogl::Shader> shader = nullptr;
		ogl::Location mvpMatrixUniform = 0;
		ogl::Location mvMatrixUniform = 0;

		auto entityView = world.find<T>();
		debug::info("Triggering rendering pass for {} entities", entityView.size());

		for (const auto &entity : entityView) {
			// ensure that the model pipeline actually exists...
			const auto &model = world.get<T>(entity);
			if (model.pipeline == nullptr) {
				debug::warning("shader was null, aborting render");
				continue;
			}

			// check if we switched shaders
			if (shader == nullptr || shader->shaderID() != model.pipeline->shaderID()) {
				// new shader - need to re-send the view and projection matrices
				shader = model.pipeline;
				shader->use();
				if (shader->hasUniform("projection")) {
					shader->setUniformMtx(shader->uniform("view"), viewMatrix);
					shader->setUniformMtx(shader->uniform("projection"), projectionMatrix);
				}
				mvpMatrixUniform = shader->uniform("MVPMatrix");
				mvMatrixUniform = shader->uniform("MVMatrix");
			}

			// set model transform
			const auto &transform = world.get<math::Transform>(entity);
			shader->setUniformMtx(mvpMatrixUniform, projectionMatrix * viewMatrix * transform.model());
			shader->setUniformMtx(mvMatrixUniform, viewMatrix * transform.model());

			auto normalMatrix = glm::mat3(glm::transpose(inverse(transform.model())));
			shader->setUniformMtx(shader->uniform("NormalMatrix"), normalMatrix);

			// setup lighting mode
			setup_lighting(shader, viewMatrix, camTransform, transform, lightPos);
			setup_material(shader, world.tryGet<PhongMaterial>(entity, &DEFAULT_MATERIAL));

			model.draw();
		}
	}

	template<typename T>
	void forward_pass_normals(const entity::EntityID& camera, const fggl::entity::EntityManager& world, const std::shared_ptr<ogl::Shader>& shader) {
		// enable required OpenGL state
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		// enable depth testing
		glEnable(GL_DEPTH_TEST);

		// set-up camera matrices
		const auto &camTransform = world.get<fggl::math::Transform>(camera);
		const auto &camComp = world.get<fggl::gfx::Camera>(camera);

		const math::mat4 projectionMatrix = camComp.perspective();
		const math::mat4 viewMatrix = glm::lookAt(camTransform.origin(), camComp.target, camTransform.up());

		ogl::Location modelUniform = shader->uniform("model");
		ogl::Location viewUniform = shader->uniform("view");
		ogl::Location projUniform = shader->uniform("projection");

		shader->use();
		shader->setUniformMtx(projUniform, projectionMatrix);
		shader->setUniformMtx(viewUniform, viewMatrix);

		auto entities = world.find<T>();
		for (const auto &entity : entities) {

			// ensure that the model pipeline actually exists...
			const auto &model = world.get<T>(entity);

			// set model transform
			const auto &transform = world.get<math::Transform>(entity);
			shader->setUniformMtx(modelUniform, transform.model());

			// render model
			model.draw();
		}
	}

	MeshData upload_mesh(const mesh::Mesh3D& meshComponent);
	std::vector<MeshData> upload_multi_mesh(const mesh::MultiMesh3D& meshComponent);

} // namespace fggl::gfx::ogl4

#endif //FGGL_GFX_OGL4_MESHES_HPP
